from asp import solve_pref, solve_cf
from itertools import combinations, chain, product
from random import random, choice, sample
import numpy as np
from scipy.special import comb
from collections import defaultdict
import pickle

def powerset(iterable):
    """powerset([1,2,3]) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)"""
    s = list(iterable)
    return chain.from_iterable(combinations(s, r) for r in range(len(s)+1))

class ABSAF:
    @classmethod
    def fromFile(cls, path):
        """ Reads an AF from an ASP-encoded file
            Assumption: Arguments are of form a1, a2, ...
                        declaring arguments: arg(a1)
                        declaring attacks:   att(a1,a2) """

        res = {}
        ASP = ""
        with open(path, "r") as file:
            for line in file.readlines():
                ASP += line
                if "arg" in line:
                    # Retains only the integers
                    argu = int("".join([c for c in line if c.isdigit()]))
                    res[argu] = []
                if "att" in line:
                    argu1, argu2 = line.split(",")
                    argu1 = int("".join([c for c in argu1 if c.isdigit()]))
                    argu2 = int("".join([c for c in argu2 if c.isdigit()]))
                    res[argu1].append(argu2)

        return cls(res)

    @classmethod
    def deserialize(cls, path):
        """ Open an ABSAF saved as pickle """
        with open(path, "rb") as f:
            af = pickle.load(f)
        return af

    def __init__(self, dictionary, representationType = "all", owaParam = 0, greedy = False, votes = None):

        """ Given a dictionary encoding an adjacency list, creates an AF. Extra arguments (each of which can be changed later on):

            representationType (str) : The strategy to compute the representation score. Can be `all` (the normal score) or `defCore`
            owaParam (int) :  The parameter of the OWA function. The OWA vector is computed as
                              [1/(1**owaParam), 1/(2**owaParam), ..., 1/(n**owaParam)]
                              For example, 0 is utilitarian, 1 is harmonic, float("inf") is egalitarian
            greedy (Bool) : whether to use the greedy functions or not
            votes : a list of frozensets of arguments.

        """

        # We assume all nodes are just given as integers.
        assert all(isinstance(node, int) for node in dictionary.keys())

        # Get the information
        self._nodes = frozenset(dictionary.keys())
        self._attacks = dictionary

        # Generate the ASP encoding
        self._inASP = ""
        for node in self._nodes:
            self._inASP += f"arg(a{node}).\n"
        self._inASP += "\n\n"

        for node, attacks in self._attacks.items():
            for attack in attacks:
                self._inASP += f"att(a{node}, a{attack}).\n"
            self._inASP += f"\n"

        # Will be filled later on; contains the preferred and CF extensions
        self._preferred = None
        self._cf = None

        self._votes = votes
        self._representationType = representationType
        self._owaParam = owaParam
        self._greedy = greedy

        # This will map each vote to the denominator needed to compute the defendable core,
        # so to avoid recomputing it.
        self._defCoreDenominator = {}

    def getVotes(self):
        return self._votes

    def setOwaValue(self, owaParam):
        # Controls the OWA function used
        # Positive integer!
        self._owaParam = owaParam

    def setRepresentationType(self, representationType):
        # Controls whether we use the normal representation, or the def-core
        assert representationType in ("all", "defCore")
        self._representationType = representationType

    def setGreedy(self, greedy):
        # Boolean value, controlling whether we use the greedy functions or not
        self._greedy = greedy

    def preferredExtensions(self):
        """ Return preferred extensions (list of frozenSets) """

        # If this entry is empty, we compute it first, once and for all, and store it

        if self._preferred is None:
            self._preferred = list(map(frozenset, solve_pref(self._inASP, 0)))  # 0: Return all extensions
        return self._preferred

    def conflictFreeExtensions(self):
        """ Return cf extensions (list of frozenSets) """

        # If this entry is empty, we compute it first, once and for all, and store it

        if self._cf is None:
            self._cf = list(map(frozenset, solve_cf(self._inASP, 0)))  # 0: Return all extensions
        return self._cf

    def _metricDistance(self, lenIntersection, lenVote):

        """ Metric distance from a groundTruth and a Vote, used when generating votes with Mallows.
            the lenIntersection is the length of the intersection between groundTruth and the vote,
            and lenVote the length of the vote. """

        return lenVote - lenIntersection

    def generateVotesMallows(self, dispersion, nGroundTruths, votersPerGroundTruth):

        """ Generate votes through our mallows method, with dispersion, nGroundTruths ground truths,
            and for each ground truth, generate votersPerGroundTruth votes.

            The result of this will be stored in the self._votes structure, but this function
            returns the groundTruths we used to generate the data (in case we need them for plotting purposes). """

        # following https://arxiv.org/pdf/2207.01140.pdf

        # If we require too many ground truths, we simply choose all pref-extensions
        if nGroundTruths >= len(self.preferredExtensions()):
            groundTruths = self.preferredExtensions()

        # Otherwise, we choose uniformly at random them
        else:
            groundTruths = sample(self.preferredExtensions(), nGroundTruths)

        # Initially empty
        self._votes = []

        # Number of arguments
        m = len(self._nodes)

        # For each ground truth, we generate some votes
        for groundTruth in groundTruths:

            # Nodes NOT in the ground truth
            externals = [a for a in self._nodes if a not in groundTruth]

            # Again, see https://arxiv.org/pdf/2207.01140.pdf for the following

            # We will first compute the probability of sampling a vote with x arguments in the ground truth,
            # and of total size y.

            f = defaultdict(float)  # we store here some intermediate value
            z = len(groundTruth)  # length of ground truth
            Z = 0  # normalization factor, see later


            values = []  # here we will store all possible values of x and y

            # Size of the intersection (x) ranges from 0 to size of groundTruth
            for x in range(len(groundTruth)+1):
                # Whereas the size of the vote (y) ranges from x to the number of arguments
                for y in range(x, m+1):

                    # First we add (x, y) to values. 
                    # Why do we encode it as the string x#y instead of the tuple (x, y)?
                    # Because the sampling function we use later doesnt like tuples, we are forced to do this.

                    assert f"{x}#{y}" not in values
                    values.append(f"{x}#{y}")

                    # (recall that metric distance only depends on x and y)
                    # Probability of computing a vote of size x and y is proportional to f(x, y),
                    # which is given by (z choose x) * (m-z choose y-x) * (dispersion**distance(x, y))
                    # (if y is 0, the probability is 0, we dont want to have empty votes)

                    f[(x, y)] = 0 if y == 0 else (comb(z, x) * comb(m-z, y-x) * (dispersion**self._metricDistance(x, y)))

                    # construct the normalization factor (sum of all f(x,y))

                    Z += f[(x, y)]

            # We now compute the probability itself
            probabilities = []
            # For every value:
            for v in values:
                # decode (x, y) from the string x#y
                x, y = map(int, v.split("#"))
                # probability is f(x, y) / sum of all f(x, y)
                probabilities.append(f[(x, y)] / Z)

            # Ok, we have the probabilities! Now we sample #votersPerGroundTruth votes!
            for _ in range(votersPerGroundTruth):
                # Chose a random set of values x#y, according to probabilities
                # (observe that values and probabilities are in the same order, so the first entry of
                # values has probability probabilities[0], etc.)
                v = np.random.choice(values, p=probabilities)
                # decode them
                x, y = map(int, v.split("#"))
                # sample x values from inside the GT
                inside = sample(groundTruth, x)
                # and y-x from outside
                outside = sample(externals, y-x)
                # make the vote, and add it
                vote = frozenset(inside + outside)
                self._votes.append(vote)

        # the votes are stored in self._votes (and accessible from self.getVotes()), but we
        # return the groundTruths here
        return groundTruths

    def computerDefCoreData(self):
        """ When called, this function computes the data
            necessary to compute the def-core-representation of every voter and stores it for later. """
        for vote in self.getVotes():
            # Calling this function once automatically computes it
            self.defCoreDenominator(vote)

    def defCoreDenominator(self, vote):

        # Returns the denominator of the def-core-representation function, given a vote

        if vote not in self._defCoreDenominator:

            # If this was never stored, store it now:
            assert type(vote) == frozenset
            self._defCoreDenominator[vote] = max(len(E.intersection(vote)) for E in self.preferredExtensions())

        return self._defCoreDenominator[vote]

    def representationVoteExt(self, vote, extension):

        """ Given an approval ballot and a set of arguments, how much does this extension represent the ballot? """

        if self._representationType == "all":  # normal representation
            return len(vote.intersection(extension)) * 1.0 / len(vote)
        elif self._representationType == "defCore":  # def-core: use the special denominator
            mu = self.defCoreDenominator(vote)
            # if the denominator is 0, return 1 (maximum representation)
            return  len(vote.intersection(extension)) * 1.0 / mu if mu > 0 else 1
        else:
            raise Exception("The possible representation types are `all` and `defCore`.")

    def representationVoteSet(self, vote, S):
        """ Given an approval ballot and a set of extensions (sets of arguments), how much does this set represent the ballot? """

        # Pick the extension that represents the ballot most (the maximum)
        return max(self.representationVoteExt(vote, extension) for extension in S)

    def representationSetOWA(self, S):
        """ Given a set of approval ballots and a set of extensions (sets of arguments), how much does this set represent the ballots?
            In this method, we use the OWA score to asses it. """

        # Sort the scores
        scores = sorted(self.representationVoteSet(vote, S) for vote in self.getVotes())
        # Compute the weights, according to the owaParam variable (use self.setOwaValue to change this!)
        weights = [1/(i**self._owaParam) for i in range(1, len(self.getVotes())+1)]
        # Return the dot product
        return sum(a*i for a, i in zip(weights, scores))

    def findRepresentationOWA(self, nExtensions):
        """ Find a set of (nExtensions)-many preferred extensions that best represents the voters, using the OWA function """

        if nExtensions >= len(self.preferredExtensions()):  # if we dont have enough, return all
            return list(self.preferredExtensions())
        else:

            best, best_score = None, -float("inf")

            if self._greedy:

                # if we are greedy, initialise the outcome as empty

                S = []

                # Until we havent got enough extensions,

                while len(S) < nExtensions:

                    # Loop through all extensions.
                    for E in self.preferredExtensions():
                        if E not in S:  # if its not already chosen

                            # Get the score of the current outcome + this extension
                            score = self.representationSetOWA(S + [E])
                            # If its the new highest score, save it 
                            if score > best_score:
                                best_score, best = score, E
                    # Extend the outcome with the best score
                    S.append(best)

                return S
            else:

                # loop through all k-combinations of extensions

                for S in combinations(self.preferredExtensions(), nExtensions):
                    # Compute score, and if it's better, save it 
                    score = self.representationSetOWA(S)
                    if score > best_score:
                        best, best_score = S, score
                return best

    def representationSetMaxCover(self, S, alpha = 1):

        """ Given a set of approval ballots and a set of extensions (sets of arguments), how much does this set represent the ballots?
            In this method, we use the (number of alpha-represented voters) score to asses it.

            By default, alpha=1."""

        return sum(1 for vote in self.getVotes() if self.representationVoteSet(vote, S) >= alpha)

    def findRepresentationMaxCover(self, nExtensions, alpha = 1):

        """ Find a set of nExtensions extensions that maximizes the alpha-represented voters.
        Note: alpha is 1 by default. """

        # This is identical to findRepresentationOWA but with a different score function
        # (representationSetMaxCover instead of representationSetOWA).
        # See the comments there.

        if nExtensions >= len(self.preferredExtensions()):
            return list(self.preferredExtensions())
        else:
            best, best_score = None, -float("inf")

            if self._greedy:
                S = []

                while len(S) < nExtensions:
                    for E in self.preferredExtensions():
                        if E not in S:
                            score = self.representationSetMaxCover(S + [E], alpha)
                            if score > best_score:
                                best_score, best = score, E
                    S.append(best)

                return S
            else:
                for S in combinations(self.preferredExtensions(), nExtensions):
                    score = self.representationSetMaxCover(S, alpha)
                    if score > best_score:
                        best, best_score = S, score
                return best


    def averageRepresentation(self, S):
        """ Given a set of extensions S, what is the average (across all voters) representation it gives? """
        
        return sum(self.representationVoteSet(vote, S) for vote in self.getVotes()) / len(self.getVotes())

    def leastRepresentation(self, S):
        """ Given a set of extensions S, what is the min (across all voters) representation it gives? """
        
        return min(self.representationVoteSet(vote, S) for vote in self.getVotes())

    def percAlphaRepresentedVoters(self, S, alpha):

        """ Given a set of extensions S, what is fraction of voters it alpha-represents? """

        return self.representationSetMaxCover(S, alpha) / len(self.getVotes())

    def serialize(self, file_name):

        """ Save this ABSAF as a pickle. """

        with open(file_name, "wb") as f:
            pickle.dump(self, f)