import os
import sys
from absaf import ABSAF
from itertools import product
from collections import defaultdict, Counter
import random
from tqdm import tqdm
import pickle
import numpy as np

# If the command line argument is unsupplied, by default we save the results
# in the experiments.pkl file.

# Setting the random seeds
random_seed = 0
os.environ['PYTHONHASHSEED']=str(random_seed)
random.seed(random_seed)
np.random.seed(random_seed)

############ Experimental setup ############

# number of ABSAF that we generate for each AF. (i.e., how many times we re-generate the votes)
ABSAFs_per_AF = 5
# Number of dispersion values to use (this is the parameter for the random distribution)
nDispersions = 11 
# Generate linearly the dispersions from 0 to 1. E.g, if nDispersions==3, the we create 0, 0.5, 1.
dispersions = np.linspace(0, 1, nDispersions)
# How many ground truths we consider when generating the votes,
# and how many voters are clustered around each.
nGroundTruths = 5
votersPerGroundTruth = 20
# These are the OWA functions we consider.
# The OWA vetors we consider are of form 1/(1**x), 1/(2**x), ..., 1/(n**x)
# Hence, 0 gives [1,...,1] (util), 1 gives [1,1/2,...,1/n] (harmonic) and infinity gives [1,0,...,0] (egalitarian)
owaValues = [0, 1, float("inf")] # util, harmonic, egal
# This is k: the number of pref-extensions we pick.
representationSize = 3
# Maximum and minimum number of preferred extensions of the AF we consider.
maxPrefExt = 10
minPrefExt = 10
# How many AFs we pick.
maxAFs = 50
# Data folder
directory = "AFs"

save_as = sys.argv[1]


############################################

# Initialize the data structures for keeping the results

# Here is how the data will be saved.
# Each dictionary (e.g., avg_acc) will be indexed by
# representation type (normal or defendable core), dispersion value, function used;
# and returns a list.

# For example, avg_acc["defCore"][0.5]["maxCover"] is a list of all
# the average-representation score (under the defCore definition) given by the outcome
# computed by maxCover (each entry in the list corresponding to a different ABSAF generated with dispersion 0.5).

avg_acc = {"all": {}, "defCore": {}}  # the average representation score 
min_acc = {"all": {}, "defCore": {}}  # the minimal representation score
recover_acc = {"all": {}, "defCore": {}}  # percentage of ground truths we recover, formally: |{E in S: E is a ground truth}| / |S|
func_score = {"all": {}, "defCore": {}}  # records the score of a function for the outcome chosen by the function. E.g., for the utilitarian score we record the sum of representation scores.
func_perfect_score = {"all": {}, "defCore": {}}  # same as above, but not of the outcome chosen by the function, but of all preferred extensions
func_greedy_score = {"all": {}, "defCore": {}}  # same as above, but not of the outcome chosen by the function, but of the greedy function
perfRepr_acc = {"all": {}, "defCore": {}}  # percentage of 1-represented voters
halfRepr_acc = {"all": {}, "defCore": {}}  # percentage of 0.5-represented voters

# Initialize the dictionaries
for representationType in ["all", "defCore"]:
    for dispersion in dispersions:
        avg_acc[representationType][dispersion] = defaultdict(list)
        min_acc[representationType][dispersion] = defaultdict(list)
        recover_acc[representationType][dispersion] = defaultdict(list)
        func_score[representationType][dispersion] = defaultdict(list)
        func_perfect_score[representationType][dispersion] = defaultdict(list)
        perfRepr_acc[representationType][dispersion] = defaultdict(list)
        halfRepr_acc[representationType][dispersion] = defaultdict(list)
        func_greedy_score[representationType][dispersion] = defaultdict(list)

# Here we will save the AFs we fetched
AFs = []

print(f"Fetching {maxAFs} `good` AFs...")

stats = Counter()  # this is just for printing info about the run

progress_bar = tqdm(total=maxAFs)  # just a visualization tool
for filename in os.listdir(directory):  # loop through the folder
    f = os.path.join(directory, filename)
    if os.path.isfile(f):
        af = ABSAF.fromFile(f)  # load the AF
        if len(af.preferredExtensions()) >= minPrefExt and len(af.preferredExtensions()) <= maxPrefExt:  # if it has the right number of extensions
            AFs.append(af)  # we keep this
            stats[len(af.preferredExtensions())] += 1
            progress_bar.update(1)
        if len(AFs) == maxAFs:  # we have enough!
            break

progress_bar.close()

print("Printing the number of AFs with certain number of pref. extensions...")
print(stats)

if len(AFs) < maxAFs:
    print(f"WARNING!!! You requested {maxAFs} `good` AFs, but I found {len(AFs)}...")


print("SETTINGS: ")
print(f"{len(AFs)} AFs, with {minPrefExt} <= #preferredExtensions <= {maxPrefExt}")
print(f"k={representationSize}, dispersions={dispersions}, {nGroundTruths} ground truths with {votersPerGroundTruth} voters each")


print(f"Doing the experiments over the {len(AFs)} fecthed AFs...")

for af in tqdm(AFs):

    # for each AF, and for each dispersion value, we generate the votes ABSAFs_per_AF times.
    # Note: product(X, Y) returns the cartesian product of X and Y.

    for t, dispersion in product(range(ABSAFs_per_AF), dispersions):

        # Generate the votes. Note that the votes will be inside
        # the af.getVotes() list. The method generateVotesMallows returns the groundTruths, not the votes;
        # we need the groundTruths later to make some statistics.
        groundTruths = af.generateVotesMallows(dispersion, nGroundTruths, votersPerGroundTruth)

        # Generate the data necessary to compute the defCore (once and for all, without doing it again and again).
        af.computerDefCoreData()

        # We test the two approaches: The "normal" and the "defendable core" representation types.
        for representationType in ["all", "defCore"]:

            # This is a global flag. All the representation scores will be computed according
            # to the normal definition (if representationType == "all") or the def core (if representationType == "defCore")
            af.setRepresentationType(representationType)

            #### we begin testing stuff now! ####

            # baseline: choose all preferred extensions

            avg_acc[representationType][dispersion]["all_pref"].append(af.averageRepresentation(af.preferredExtensions()))
            min_acc[representationType][dispersion]["all_pref"].append(af.leastRepresentation(af.preferredExtensions()))
            recover_acc[representationType][dispersion]["all_pref"].append(1.0) #  this is always perfect
            perfRepr_acc[representationType][dispersion]["all_pref"].append(af.percAlphaRepresentedVoters(af.preferredExtensions(), 1.0))
            halfRepr_acc[representationType][dispersion]["all_pref"].append(af.percAlphaRepresentedVoters(af.preferredExtensions(), 0.5))

            # baseline: choose k random preferred extensions

            S = random.sample(af.preferredExtensions(), representationSize)
            avg_acc[representationType][dispersion]["random"].append(af.averageRepresentation(S))
            min_acc[representationType][dispersion]["random"].append(af.leastRepresentation(S))
            recover_acc[representationType][dispersion]["random"].append(sum(1.0 for E in S if E in groundTruths) / len(S))
            perfRepr_acc[representationType][dispersion]["random"].append(af.percAlphaRepresentedVoters(S, 1.0))
            halfRepr_acc[representationType][dispersion]["random"].append(af.percAlphaRepresentedVoters(S, 0.5))

            # baseline: choose all ground truths
            S = groundTruths
            avg_acc[representationType][dispersion]["all_gt"].append(af.averageRepresentation(S))
            min_acc[representationType][dispersion]["all_gt"].append(af.leastRepresentation(S))
            recover_acc[representationType][dispersion]["all_gt"].append(sum(1.0 for E in S if E in groundTruths) / len(S))
            perfRepr_acc[representationType][dispersion]["all_gt"].append(af.percAlphaRepresentedVoters(S, 1))   
            halfRepr_acc[representationType][dispersion]["all_gt"].append(af.percAlphaRepresentedVoters(S, 0.5))   

            # Enough baselines. Now we use our methods.
            # In particular, we first try all the OWA functions.

            for owaValue in owaValues:

                # Global flag that sets the OWA function to use.
                # For example, if owaValue == 0, we will be using the utilitarian version

                af.setOwaValue(owaValue)

                # Global flag: We use the greedy-OWA first.

                af.setGreedy(True)

                # compute the greedyOWA outcome of size k

                S = af.findRepresentationOWA(representationSize)
                func_greedy_score[representationType][dispersion][owaValue].append(af.representationSetOWA(S))

                # non-greedy now.

                af.setGreedy(False)

                S = af.findRepresentationOWA(representationSize)
                                    
                avg_acc[representationType][dispersion][owaValue].append(af.averageRepresentation(S))
                min_acc[representationType][dispersion][owaValue].append(af.leastRepresentation(S))
                func_score[representationType][dispersion][owaValue].append(af.representationSetOWA(S))
                func_perfect_score[representationType][dispersion][owaValue].append(af.representationSetOWA(af.preferredExtensions()))
                recover_acc[representationType][dispersion][owaValue].append(sum(1.0 for E in S if E in groundTruths) / len(S))
                perfRepr_acc[representationType][dispersion][owaValue].append(af.percAlphaRepresentedVoters(S, 1.0))
                halfRepr_acc[representationType][dispersion][owaValue].append(af.percAlphaRepresentedVoters(S, 0.5))

            ## END OF OWA PART: now we try the maxCover function. Again, greedy first.

            af.setGreedy(True)

            S = af.findRepresentationMaxCover(representationSize)
            func_greedy_score[representationType][dispersion]["maxCover"].append(af.representationSetMaxCover(S))

            # Non-greedy.

            af.setGreedy(False)

            S = af.findRepresentationMaxCover(representationSize)
                                
            avg_acc[representationType][dispersion]["maxCover"].append(af.averageRepresentation(S))
            min_acc[representationType][dispersion]["maxCover"].append(af.leastRepresentation(S))
            func_score[representationType][dispersion]["maxCover"].append(af.representationSetMaxCover(S))
            func_perfect_score[representationType][dispersion]["maxCover"].append(af.representationSetMaxCover(af.preferredExtensions()))
            recover_acc[representationType][dispersion]["maxCover"].append(sum(1.0 for E in S if E in groundTruths) / len(S))
            perfRepr_acc[representationType][dispersion]["maxCover"].append(af.percAlphaRepresentedVoters(S, 1.0))
            halfRepr_acc[representationType][dispersion]["maxCover"].append(af.percAlphaRepresentedVoters(S, 0.5))

    
    # We store all the resulting data inside this dictionary, plus some info
    # (i,e., the dispersions used, and all functions considered (including baselines))
    # and save it as a pickle (e.g., we save it as it is, as an object. This
    # is non-risky, as we are saving a dictionary/list-based structure.)

    # Note: We save this at every iteration... that is,
    # we can analyze the results via plot_experiments.py during execution!
    results = {
        "avg" : avg_acc,
        "min" : min_acc,
        "recover" : recover_acc,
        "score" : func_score,
        "perfect" : func_perfect_score,
        "perfRepr" : perfRepr_acc,
        "halfRepr" : halfRepr_acc,
        "greedy" : func_greedy_score,
        "dispersions" : dispersions,
        "functions" : owaValues + ["maxCover", "all_pref", "random", "all_gt"]
    }

    # Save

    with open(f"{save_as}", "wb") as f:
        pickle.dump(results, f)