import os
import sys
from absaf import ABSAF
import random
from tqdm import tqdm
import pickle
import numpy as np
from collections import Counter
from itertools import product
import multiprocessing
from time import time


# Setting the random seeds
random_seed = 0
os.environ['PYTHONHASHSEED']=str(random_seed)
random.seed(random_seed)
np.random.seed(random_seed)

############ Experimental setup ############

# Number of dispersion values to use (this is the parameter for the random distribution)
nDispersions = 11 
# Generate linearly the dispersions from 0 to 1. E.g, if nDispersions==3, the we create 0, 0.5, 1.
dispersions = np.linspace(0, 1, nDispersions)
# Size of preferred extensions that we start the analysis from: {basePrefExt-2 , ... , basePrefExt+2}
basePrefExt = 10
jump = 2
# How many AFs we pick per size-group.
maxAFs = 30
# approximate number of total voters
approxNVoters = 100
# Data folder
directory = "AFs"

function = "util" # util, egal, or maxCov

timeout = 45

save_as = sys.argv[1]

############################################

# Initialize the data structures for keeping the results

counter_base = {}
counter_greedy = {}

times_base = {}
times_greedy = {}

# Size of extensions to look for
minPrefExt = basePrefExt - jump
maxPrefExt = basePrefExt + jump

while True:

    print("Collecting...")

    # Here we will save the AFs we fetched
    AFs = []

    for filename in os.listdir(directory):  # loop through the folder
        f = os.path.join(directory, filename)
        if os.path.isfile(f):
            af = ABSAF.fromFile(f)  # load the AF
            if len(af.preferredExtensions()) >= minPrefExt and len(af.preferredExtensions()) <= maxPrefExt:  # if it has the right number of extensions
                AFs.append(af)  # we keep this
            if len(AFs) == maxAFs:  # we have enough!
                break

    assert len(AFs) >= maxAFs

    # Initialize data structure for this size batch

    counter_base[(minPrefExt, maxPrefExt)] = 0 # stores the timesouts of base
    counter_greedy[(minPrefExt, maxPrefExt)] = 0 # stores the timesouts of greedy

    times_base[(minPrefExt, maxPrefExt)] = [] # stores the average runtime of successes of base
    times_greedy[(minPrefExt, maxPrefExt)] = [] # stores the average runtime of successes of greedy

    print(minPrefExt, maxPrefExt)

    # For each AF
    for af in tqdm(AFs):

        # k is circa  |pref(AF)|/4
        k = len(af.preferredExtensions()) // 4
        nGroundTruths = k

        votersPerGroundTruth = (approxNVoters // nGroundTruths)

        # We pick a random dispersion (not important here)
        af.generateVotesMallows(random.choice(dispersions), nGroundTruths, votersPerGroundTruth)

        # nongreedy

        af.setGreedy(False)


        # Run the rule as a subprocess

        if function == "util":
            af.setOwaValue(0)
            p = multiprocessing.Process(target=af.findRepresentationOWA, name="MyThing", args=(k,))
        elif function == "egal":
            af.setOwaValue(float("inf"))
            p = multiprocessing.Process(target=af.findRepresentationOWA, name="MyThing", args=(k,))
        elif function == "maxCov":
            p = multiprocessing.Process(target=af.findRepresentationMaxCover, name="MyThing", args=(k,))

        start = time()
                
        p.start()

        p.join(timeout)

        if p.is_alive():
            # Timeout: kill the process
            p.terminate()
            p.join()
        else:
            # Success!
            counter_base[(minPrefExt, maxPrefExt)] += 1
            times_base[(minPrefExt, maxPrefExt)].append(time() - start)

        # greedy, similar code

        af.setGreedy(True)

        if function == "util":
            af.setOwaValue(0)
            p = multiprocessing.Process(target=af.findRepresentationOWA, name="MyThing", args=(k,))
        elif function == "egal":
            af.setOwaValue(float("inf"))
            p = multiprocessing.Process(target=af.findRepresentationOWA, name="MyThing", args=(k,))
        elif function == "maxCov":
            p = multiprocessing.Process(target=af.findRepresentationMaxCover, name="MyThing", args=(k,))

        start = time()

        p.start()

        p.join(timeout)

        if p.is_alive():
            p.terminate()
            p.join()
        else:
            counter_greedy[(minPrefExt, maxPrefExt)] += 1
            times_greedy[(minPrefExt, maxPrefExt)].append(time() - start)

    # Print the results
    print(f"base: {counter_base[(minPrefExt, maxPrefExt)]}/{maxAFs} successes" + (f", average time of successes {np.mean(times_base[(minPrefExt, maxPrefExt)]):.4f}s" if times_base[(minPrefExt, maxPrefExt)] else ""))
    print(f"greedy: {counter_greedy[(minPrefExt, maxPrefExt)]}/{maxAFs} successes" + (f", average time of successes {np.mean(times_greedy[(minPrefExt, maxPrefExt)]):.4f}s" if times_greedy[(minPrefExt, maxPrefExt)] else ""))
    if times_base[(minPrefExt, maxPrefExt)] and times_greedy[(minPrefExt, maxPrefExt)]:
        print(f"improvement ratio: {np.mean(times_base[(minPrefExt, maxPrefExt)]):.4f}s/{np.mean(times_greedy[(minPrefExt, maxPrefExt)]):.4f}s = {np.mean(times_base[(minPrefExt, maxPrefExt)])/np.mean(times_greedy[(minPrefExt, maxPrefExt)]):.4f}x")

    print()

    # save results

    results = {
        "counter_base" : counter_base,
        "counter_greedy" : counter_greedy,
        "times_base" : times_base,
        "times_greedy" : times_greedy,
        "function" : function
    }

    with open(f"{save_as}", "wb") as f:
        pickle.dump(results, f)

    # If we never succeeded in the non-greedy version, stop.

    if counter_base[(minPrefExt, maxPrefExt)] == 0:
        assert counter_greedy[(minPrefExt, maxPrefExt)] > 0
        break

    # Increase size-group

    minPrefExt = maxPrefExt+1
    maxPrefExt = minPrefExt + 2*jump