import pickle
import numpy as np
import matplotlib.pyplot as plt
import sys

# If the command line argument is unsupplied, by default we plot the results
# in the experiments.pkl file.

file = sys.argv[1]

with open(file, "rb") as f:
    results = pickle.load(f)

# First, we get the dispersions and functions considered
dispersions = results["dispersions"]
functions = results["functions"]

# This is for plotting reasons,
# it is the x-axis (corresponds to the number of dispersions used)
ticks = range(len(dispersions))

# Again for visualization purposes.
# It matches the nicknames used in the experiments
# to a proper name
# e.g., 0 (owaValue 0) corresponds to utilitarian,
# all_pref to all preferrex extension

name = {
    0: "Utilitarian",
    0.5: "1/sqrt(x)",
    1: "Harmonic",
    2: "1/(x^2)",
    float("inf"): "Egalitarian",
    "maxCover": "MaxCover",
    "all_pref": "all pref.",
    "random": "k random extensions",
    "all_gt": "all ground truths"
}

metric_name = {
    "avg": "Average representation",
    "min": "Minimum representation",
    "recover": "Recovery score",
    "perfRepr": "Perfect representation score",
    "halfRepr": "halfRepr"
}

# Generate the plots. 
# For representation type and metric,

for representationType in ["all", "defCore"]:
    # average and minimal repr. score, ground-truth-recover score, % of 1-represented, % of 0.5 represented
    for metric in ["avg", "min", "recover", "perfRepr", "halfRepr"]:

        # we make a plot. We plot all functions:

        for f in functions:

            # recall that
            # results[metric][representationType][dispersion][f]
            # is a list of all the metric-score (e.g., average-representation scores)
            # under representationType (e.g., defendable core)
            # given by the set chosen by function f (e.g., utilitarian)
            # for the ABSAFs sampled with this dispersion (e.g. 0.5)

            # we aggregate all results per dispersion, and average

            res = []
            for dispersion in dispersions:
                res.append(np.mean(results[metric][representationType][dispersion][f]))

            # plot this function

            plt.plot(ticks, res, label=name[f])

        # ok, plotted al functions: Now we show the plot

        # Adding labels and title
        plt.xlabel('dispersion value')
        plt.ylabel(metric_name[metric])
        plt.title("representation: " + ("Regular" if representationType == "all" else "Defendable Core"))

        # Adding a legend to identify each curve
        plt.legend()

        # axis: add the dispersions value
        plt.xticks(ticks, (f"{dispersion:.1f}" for dispersion in dispersions))

        # Display the plot
        plt.show()

### Second part of the experiment: No metric, just approximation values

# Therefore, we remove the baselines....

for f in ("all_pref", "all_gt", "random"):
    if f in functions:
        functions.remove(f)

# Here, ideal is the score we want to approximate,
# and approximator the function we are judging. For example,
# if ideal == perfect and approximator == score, then 
# we are showing the approximation-ratio of a function (say, utilitarian)
# vs the utilitarian score of using all preferred extensions (that's the ideal)

# On the other hand, if ideal == score and approximator == greedy, then 
# we are showing the approximation-ratio of a greedy-function (say, greedy-utilitarian)
# vs the utilitarian score of using the normal utilitarian function

for ideal, approximator in [("perfect", "score"), ("score", "greedy")]:
    for representationType in ["all", "defCore"]:
        for f in functions:

            # Inner loop similar to above... 

            res = []
            for dispersion in dispersions:
                tmp = []  # will contain the result

                # results[ideal][representationType][dispersion][f] and results[approximator][representationType][dispersion][f]
                # are lists of scores (each entry of this list being an individual ABSAF)
                # we loop through them together:


                for ideal_score, approximator_score in zip(results[ideal][representationType][dispersion][f], results[approximator][representationType][dispersion][f]):
                    # If the ideal score is 0, then 
                    if ideal_score == 0:
                        # sanity check: the score has to be 0 as well (can't be higher)
                        assert ideal_score == approximator_score
                        # and although degenerate this is a 100% approximation
                        tmp.append(100)

                    # Otherwise, compute percentage
                    else:
                        tmp.append(approximator_score * 100 / ideal_score)

                # compute average of this approximation ratios
                res.append(np.mean(tmp))

            # Plot results for this function

            plt.plot(ticks, res, label=name[f])

        # Visualize the plots!

        # Adding labels and title
        plt.xlabel('dispersion')
        plt.ylabel(f"percentage of {ideal.upper()} score via {approximator.upper()}")
        plt.title("representation: " + ("Regular" if representationType == "all" else "Defendable Core"))

        # Adding a legend to identify each curve
        plt.legend()

        # axis
        plt.xticks(ticks, (f"{dispersion:.1f}" for dispersion in dispersions))

        # Display the plot
        plt.show()