import os
import sys
from absaf import ABSAF
import random
from tqdm import tqdm
import pickle
import numpy as np
from collections import Counter
from itertools import product

# If the command line argument is unsupplied, by default we save the results
# in the experiments.pkl file.

save_as = sys.argv[1]

# Setting the random seeds
random_seed = 0
os.environ['PYTHONHASHSEED']=str(random_seed)
random.seed(random_seed)
np.random.seed(random_seed)

############ Experimental setup ############

# number of ABSAF that we generate for each AF. (i.e., how many times we re-generate the votes)
ABSAFs_per_AF = 5
# Number of dispersion values to use (this is the parameter for the random distribution)
nDispersions = 11 
# Generate linearly the dispersions from 0 to 1. E.g, if nDispersions==3, the we create 0, 0.5, 1.
dispersions = np.linspace(0, 1, nDispersions)
# How many ground truths we consider when generating the votes,
# and how many voters are clustered around each.
nGroundTruths = 5
votersPerGroundTruth = 20
# This is the ks we consider: the number of pref-extensions we pick.
representationSizes = [1, 2, 3, 4, 5, 6, 7]
# Maximum and minimum number of preferred extensions of the AF we consider.
maxPrefExt = 10
minPrefExt = 10
# How many AFs we pick.
maxAFs = 50
# Data folder
directory = "AFs"

# To run experiments in. In the main body of paper, we just show utilitarian
functions = ["util", "harmonic", "egal", "greedy_harmonic", "greedy_util", "greedy_egal"]

############################################

# Initialize the data structures for keeping the results

func_score = {}
for function in functions:
    func_score[function] = {}
    for representationSize in representationSizes:
        func_score[function][representationSize] = {}
        for representationType in ["all", "defCore"]:
            func_score[function][representationSize][representationType] = {}
            for dispersion in dispersions:
                func_score[function][representationSize][representationType][dispersion] = []

# Here we will save the AFs we fetched
AFs = []

print(f"Fetching {maxAFs} `good` AFs...")

stats = Counter()  # this is just for printing info about the run

progress_bar = tqdm(total=maxAFs)  # just a visualization thing
for filename in os.listdir(directory):  # loop through the folder
    f = os.path.join(directory, filename)
    if os.path.isfile(f):
        af = ABSAF.fromFile(f)  # load the AF
        if len(af.preferredExtensions()) >= minPrefExt and len(af.preferredExtensions()) <= maxPrefExt:  # if it has the right number of extensions
            AFs.append(af)  # we keep this
            stats[len(af.preferredExtensions())] += 1
            progress_bar.update(1)
        if len(AFs) == maxAFs:  # we have enough!
            break

progress_bar.close()

print("Printing the number of AFs with certain number of pref. extensions...")
print(stats)

if len(AFs) < maxAFs:
    print(f"WARNING!!! You requested {maxAFs} `good` AFs, but I found {len(AFs)}...")

print("SETTINGS: ")
print(f"{len(AFs)} AFs, with {minPrefExt} <= #preferredExtensions <= {maxPrefExt}")
print(f"k={representationSizes}, dispersions={dispersions}, {nGroundTruths} ground truths with {votersPerGroundTruth} voters each")

print(f"Doing the experiments over the {len(AFs)} fecthed AFs...")



for af in tqdm(AFs):

    # for each AF, and for each dispersion value, we generate the votes ABSAFs_per_AF times.
    # Note: product(X, Y) returns the cartesian product of X and Y.

    for t, dispersion in product(range(ABSAFs_per_AF), dispersions):

        # Generate the votes. Note that the votes will be inside
        # the af.getVotes() list. The method generateVotesMallows returns the groundTruths, not the votes;
        # we need the groundTruths later to make some statistics.
        groundTruths = af.generateVotesMallows(dispersion, nGroundTruths, votersPerGroundTruth)

        # Generate the data necessary to compute the defCore (once and for all, without doing it again and again).
        af.computerDefCoreData()

        # We test the two approaches: The "normal" and the "defendable core" representation types.
        for representationType in ["all", "defCore"]:

            # This is a global flag. All the representation scores will be computed according
            # to the normal definition (if representationType == "all") or the def core (if representationType == "defCore")
            af.setRepresentationType(representationType)

            for representationSize in representationSizes:

                # For each function...

                for function in functions:

                    # Compute the outcome, check its score, and store it in the result

                    if function == "util":
                        af.setOwaValue(0)
                        af.setGreedy(False)
                        S = af.findRepresentationOWA(representationSize)
                        func_score[function][representationSize][representationType][dispersion].append(af.representationSetOWA(S))
                    elif function == "greedy_util":
                        af.setOwaValue(0)
                        af.setGreedy(True)
                        S = af.findRepresentationOWA(representationSize)
                        func_score[function][representationSize][representationType][dispersion].append(af.representationSetOWA(S))
                    elif function == "harmonic":
                        af.setOwaValue(1)
                        af.setGreedy(False)
                        S = af.findRepresentationOWA(representationSize)
                        func_score[function][representationSize][representationType][dispersion].append(af.representationSetOWA(S))
                    elif function == "greedy_harmonic":
                        af.setOwaValue(1)
                        af.setGreedy(True)
                        S = af.findRepresentationOWA(representationSize)
                        func_score[function][representationSize][representationType][dispersion].append(af.representationSetOWA(S))
                    elif function == "egal":
                        af.setOwaValue(float("inf"))
                        af.setGreedy(False)
                        S = af.findRepresentationOWA(representationSize)
                        func_score[function][representationSize][representationType][dispersion].append(af.representationSetOWA(S))
                    elif function == "greedy_egal":
                        af.setOwaValue(float("inf"))
                        af.setGreedy(True)
                        S = af.findRepresentationOWA(representationSize)
                        func_score[function][representationSize][representationType][dispersion].append(af.representationSetOWA(S))
                    else:
                        raise NotImplementedError()
    
    # We store all the resulting data inside this dictionary, plus some info
    # (i,e., the dispersions used, the various values of k, and all functions considered)
    # and save it as a pickle (e.g., we save it as it is, as an object. This
    # is non-risky, as we are saving a dictionary/list-based structure.)

    # Note: We save this at every iteration... that is,
    # we can analyze the results via plot_varying_k.py during execution!
    results = {
        "dispersions" : dispersions,
        "functions" : functions,
        "representationSizes": representationSizes
    }

    for function in functions:
        results[function] = func_score[function]

    # Save

    with open(f"{save_as}", "wb") as f:
        pickle.dump(results, f)